'use strict';

const fs = require('node:fs/promises');
const fs$1 = require('node:fs');
const path = require('node:path');
const postcss = require('postcss');
const atImport = require('postcss-import');
const t = require('@babel/types');
const babelGenerate = require('@babel/generator');
const node_url = require('node:url');

function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e.default : e; }

function _interopNamespaceCompat(e) {
  if (e && typeof e === 'object' && 'default' in e) return e;
  const n = Object.create(null);
  if (e) {
    for (const k in e) {
      n[k] = e[k];
    }
  }
  n.default = e;
  return n;
}

const fs__default = /*#__PURE__*/_interopDefaultCompat(fs);
const fs__default$1 = /*#__PURE__*/_interopDefaultCompat(fs$1);
const path__default = /*#__PURE__*/_interopDefaultCompat(path);
const postcss__default = /*#__PURE__*/_interopDefaultCompat(postcss);
const atImport__default = /*#__PURE__*/_interopDefaultCompat(atImport);
const t__namespace = /*#__PURE__*/_interopNamespaceCompat(t);
const babelGenerate__default = /*#__PURE__*/_interopDefaultCompat(babelGenerate);

const markedLayerKey = Symbol("__tw_layer_name__");
const layerNodesKeys = ["base", "components", "utilities"];

const atRulesNameFilter = "twlayer";
const atRulesRenamePlugin = () => {
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-at-rule-rename-plugin",
    Once(root) {
      root.walkAtRules("layer", (rule) => {
        rule.name = atRulesNameFilter;
      });
    }
  };
};
atRulesRenamePlugin.postcss = true;
const markLayerPlugin = (options) => {
  const { ctx } = options;
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-mark-layer-plugin",
    Once(root) {
      root.walkAtRules(atRulesNameFilter, (rule) => {
        if (layerNodesKeys.includes(rule.params)) {
          const layerName = rule.params;
          for (const node of rule.nodes) {
            Object.defineProperty(node, markedLayerKey, {
              value: layerName,
              enumerable: false,
              configurable: true
            });
          }
          rule.parent?.insertBefore(rule, rule.nodes);
          rule.remove();
        }
      });
      if (ctx.options.outSideLayerCss !== void 0) {
        root.walkRules((rule) => {
          const layerName = rule[markedLayerKey];
          if (!layerName) {
            Object.defineProperty(rule, markedLayerKey, {
              value: ctx.options.outSideLayerCss,
              enumerable: false,
              configurable: true
            });
          }
        });
      }
    }
  };
};
markLayerPlugin.postcss = true;
const extractLayerPlugin = (options) => {
  const { ctx } = options;
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-extract-layer-plugin",
    OnceExit(root) {
      root.walkRules((rule) => {
        const layerName = rule[markedLayerKey];
        if (layerName && layerNodesKeys.includes(layerName)) {
          ctx.append(layerName, rule);
        }
      });
      if (ctx.options?.interceptors?.css)
        for (const fn of ctx.options.interceptors.css) {
          fn(root, ctx);
        }
    }
  };
};
extractLayerPlugin.postcss = true;

function tryLoadSass() {
  return import('sass');
}
function tryLoadSassSync() {
  return require("sass");
}
function isExtSassFile(extname) {
  return extname === ".scss" || extname === ".sass";
}
async function sassCompile(path, sassOptions) {
  const { compile } = await tryLoadSass();
  const { css } = compile(path, sassOptions);
  return css;
}
function sassCompileSync(path, sassOptions) {
  const { compile } = tryLoadSassSync();
  const { css } = compile(path, sassOptions);
  return css;
}

const expandAST = ["addBase", "addComponents", "addUtilities", "theme", "addVariant", "config", "corePlugins", "e", "matchComponents", "matchUtilities", "matchVariant"].map(
  (x) => {
    return t__namespace.objectProperty(t__namespace.identifier(x), t__namespace.identifier(x));
  }
);
function expandAPI() {
  return expandAST;
}
function unwrapThemeFunctionArg(str) {
  if (str) {
    const start = str[0];
    const last = str.at(-1);
    if (start === '"' && last === '"' || start === "'" && last === "'") {
      return str.slice(1, -1);
    }
  }
  return str;
}
function makeObjectProperty(key, value) {
  return t__namespace.objectProperty(t__namespace.stringLiteral(key), value);
}
function addSuffix(value, important) {
  if (typeof value === "string") {
    return value + (important ? " !important" : "");
  }
  return value;
}
function makeObjectExpression(nodes) {
  const res = nodes.reduce((acc, _cur) => {
    if (_cur.type === "rule") {
      const cur = _cur;
      const key = cur.selector;
      const value = cur.nodes.reduce((cacc, ccur) => {
        if (ccur.type === "decl") {
          const v = ccur.value;
          const important = ccur.important;
          const prop = ccur.prop;
          const arr = [...v.matchAll(/theme\(([^()]+?)\)/g)];
          if (arr.length > 0) {
            const first = arr[0];
            if (first[0] === v) {
              const ccc = t__namespace.callExpression(t__namespace.identifier("theme"), [t__namespace.stringLiteral(unwrapThemeFunctionArg(first[1]))]);
              cacc[prop] = important ? t__namespace.binaryExpression("+", ccc, t__namespace.stringLiteral(" !important")) : ccc;
            } else {
              let p = 0;
              const quasis = [];
              const expressions = [];
              for (let i = 0; i < arr.length; i++) {
                const hit = arr[i];
                if (typeof hit.index === "number") {
                  const raw = v.slice(p, hit.index);
                  p = hit.index + hit[0].length;
                  quasis.push(t__namespace.templateElement({ raw }));
                  expressions.push(t__namespace.callExpression(t__namespace.identifier("theme"), [t__namespace.stringLiteral(unwrapThemeFunctionArg(hit[1]))]));
                  if (i === arr.length - 1) {
                    quasis.push(t__namespace.templateElement({ raw: addSuffix(v.slice(p), important) }));
                  }
                }
              }
              cacc[prop] = t__namespace.templateLiteral(quasis, expressions);
            }
          } else {
            cacc[prop] = t__namespace.stringLiteral(addSuffix(v, important));
          }
        }
        return cacc;
      }, {});
      if (acc[key]) {
        acc[key].push(
          ...Object.entries(value).map(([prop, value2]) => {
            return makeObjectProperty(prop, value2);
          })
        );
      } else {
        acc[key] = Object.entries(value).map(([prop, value2]) => {
          return makeObjectProperty(prop, value2);
        });
      }
    }
    return acc;
  }, {});
  return Object.entries(res).map(([key, value]) => {
    return t__namespace.objectProperty(
      t__namespace.stringLiteral(key),
      t__namespace.objectExpression(
        Object.values(
          value.reduce((acc, cur) => {
            if (cur.key.type === "StringLiteral") {
              acc[cur.key.value] = cur;
            }
            return acc;
          }, {})
        )
      )
    );
  });
}
function getVarName(layer) {
  return `_${layer}Css`;
}
const pluginNameMap = {
  base: "addBase",
  components: "addComponents",
  utilities: "addUtilities"
};
function getFnName(key) {
  return pluginNameMap[key];
}
function getLegacyStatement(key, ctx) {
  return t__namespace.expressionStatement(t__namespace.callExpression(t__namespace.identifier(getFnName(key)), [t__namespace.objectExpression(makeObjectExpression(ctx.getNodes(key)))]));
}
const returnSelfArrowFnAst = t__namespace.arrowFunctionExpression([t__namespace.identifier("x")], t__namespace.identifier("x"));
const declReturnSelfArrowFnName = "returnSelfNoop";
const declReturnSelfArrowFnAst = t__namespace.variableDeclaration("const", [t__namespace.variableDeclarator(t__namespace.identifier(declReturnSelfArrowFnName), returnSelfArrowFnAst)]);
function createGenerator() {
  const callFnId = "_plugin";
  const pluginName = "css2TwPlugin";
  const withOptionsParamsId = "_options";
  const traverseMethodName = "withOptionsWalkCSSRuleObject";
  const requireStatement = t__namespace.variableDeclaration("const", [
    t__namespace.variableDeclarator(t__namespace.identifier(callFnId), t__namespace.callExpression(t__namespace.identifier("require"), [t__namespace.stringLiteral("tailwindcss/plugin")]))
  ]);
  const exportsStatement = t__namespace.expressionStatement(t__namespace.assignmentExpression("=", t__namespace.memberExpression(t__namespace.identifier("module"), t__namespace.identifier("exports")), t__namespace.identifier(pluginName)));
  function getLatestStatement(key, ctx) {
    const varName = getVarName(key);
    const fnName = getFnName(key);
    return [
      t__namespace.variableDeclaration("const", [
        t__namespace.variableDeclarator(
          t__namespace.identifier(varName),
          t__namespace.callExpression(t__namespace.identifier(traverseMethodName), [t__namespace.objectExpression(makeObjectExpression(ctx.getNodes(key))), t__namespace.stringLiteral(key)])
        )
      ]),
      t__namespace.expressionStatement(t__namespace.callExpression(t__namespace.identifier(fnName), [t__namespace.identifier(varName)]))
    ];
  }
  function generate(ctx, opt) {
    function getPluginBody() {
      const statements = ctx.options.withOptions ? layerNodesKeys.flatMap((layer) => {
        return getLatestStatement(layer, ctx);
      }) : layerNodesKeys.map((layer) => {
        return getLegacyStatement(layer, ctx);
      });
      const innerPluginAst = t__namespace.functionExpression(
        null,
        [t__namespace.objectPattern(expandAPI())],
        t__namespace.blockStatement(statements)
      );
      if (ctx.options.withOptions) {
        return t__namespace.callExpression(t__namespace.memberExpression(t__namespace.identifier(callFnId), t__namespace.identifier("withOptions")), [
          t__namespace.functionExpression(
            null,
            [t__namespace.assignmentPattern(t__namespace.identifier(withOptionsParamsId), t__namespace.objectExpression([]))],
            t__namespace.blockStatement([
              // const { withOptionsWalkCSSRuleObject = (x)=>x } = _options
              t__namespace.variableDeclaration("const", [
                t__namespace.variableDeclarator(
                  t__namespace.objectPattern([
                    t__namespace.objectProperty(t__namespace.identifier(traverseMethodName), t__namespace.assignmentPattern(t__namespace.identifier(traverseMethodName), t__namespace.identifier(declReturnSelfArrowFnName)))
                  ]),
                  t__namespace.identifier(withOptionsParamsId)
                )
              ]),
              t__namespace.returnStatement(innerPluginAst)
            ])
          ),
          t__namespace.functionExpression(null, [t__namespace.identifier(withOptionsParamsId)], t__namespace.blockStatement([t__namespace.returnStatement(t__namespace.objectExpression([]))]))
        ]);
      }
      return t__namespace.callExpression(t__namespace.identifier(callFnId), [innerPluginAst]);
    }
    const ast = t__namespace.file(
      t__namespace.program([requireStatement, declReturnSelfArrowFnAst, t__namespace.variableDeclaration("const", [t__namespace.variableDeclarator(t__namespace.identifier(pluginName), getPluginBody())]), exportsStatement])
    );
    const res = babelGenerate__default(ast, opt);
    return res.code;
  }
  return {
    generate
  };
}

function isObject(value) {
  return value !== null && typeof value === "object";
}
function _defu(baseObject, defaults, namespace = ".", merger) {
  if (!isObject(defaults)) {
    return _defu(baseObject, {}, namespace, merger);
  }
  const object = Object.assign({}, defaults);
  for (const key in baseObject) {
    if (key === "__proto__" || key === "constructor") {
      continue;
    }
    const value = baseObject[key];
    if (value === null || value === void 0) {
      continue;
    }
    if (merger && merger(object, key, value, namespace)) {
      continue;
    }
    if (Array.isArray(value) && Array.isArray(object[key])) {
      object[key] = [...value, ...object[key]];
    } else if (isObject(value) && isObject(object[key])) {
      object[key] = _defu(
        value,
        object[key],
        (namespace ? `${namespace}.` : "") + key.toString(),
        merger
      );
    } else {
      object[key] = value;
    }
  }
  return object;
}
function createDefu(merger) {
  return (...arguments_) => (
    // eslint-disable-next-line unicorn/no-array-reduce
    arguments_.reduce((p, c) => _defu(p, c, "", merger), {})
  );
}
const defu = createDefu();

function getDefaultCacheDir() {
  return path__default.resolve(process.cwd(), "node_modules", ".css-to-tailwindcss-plugin");
}
function getDefaults() {
  const pkgPath = require.resolve("tailwindcss/package.json");
  const nodeModulesPath = path__default.dirname(path__default.dirname(pkgPath));
  return {
    atImportOptions: {},
    sassOptions: {
      importers: [
        {
          // https://github.com/webpack-contrib/sass-loader/blob/master/src/index.js#L55
          findFileUrl(url) {
            const twUrl = node_url.pathToFileURL(path__default.resolve(nodeModulesPath, url));
            return new URL(twUrl);
          }
        }
      ]
    },
    tailwindcssResolved: false,
    withOptions: true,
    withOptionsWalkCSSRuleObject(obj) {
      return obj;
    },
    cacheDir: getDefaultCacheDir()
  };
}
function getOptions(opts) {
  return defu(opts, getDefaults());
}

function createContext(opts) {
  const options = getOptions(opts);
  const layersMap = /* @__PURE__ */ new Map();
  function resetLayersMap() {
    layersMap.set("base", []);
    layersMap.set("components", []);
    layersMap.set("utilities", []);
  }
  resetLayersMap();
  function getNodes(key) {
    const arrRef = layersMap.get(key);
    if (Array.isArray(arrRef)) {
      return arrRef;
    }
    const arr = [];
    layersMap.set(key, arr);
    return arr;
  }
  function append(key, nodes) {
    const arr = getNodes(key);
    if (Array.isArray(nodes)) {
      arr.push(...nodes);
    } else {
      arr.push(nodes);
    }
  }
  const generator = createGenerator();
  return {
    options,
    layersMap,
    append,
    getNodes,
    // not
    getPluginsSync() {
      const opt = { ctx: this };
      const plugins = [atRulesRenamePlugin()];
      plugins.push(markLayerPlugin(opt));
      plugins.push(extractLayerPlugin(opt));
      options.postcssPlugins?.(plugins);
      return plugins;
    },
    async getPlugins() {
      const { tailwindcssConfig, tailwindcssResolved } = options;
      const opt = { ctx: this };
      const plugins = [atRulesRenamePlugin()];
      if (tailwindcssResolved && tailwindcssConfig) {
        const { default: tailwindcss } = await import('tailwindcss');
        plugins.push(tailwindcss(tailwindcssConfig));
      }
      plugins.push(markLayerPlugin(opt));
      plugins.push(extractLayerPlugin(opt));
      options.postcssPlugins?.(plugins);
      return plugins;
    },
    async process(entry) {
      const { sassOptions, atImportOptions } = options;
      const plugins = await this.getPlugins();
      let css;
      const extname = path__default.extname(entry);
      if (isExtSassFile(extname)) {
        try {
          css = await sassCompile(entry, sassOptions);
        } catch (error) {
          throw error;
        }
      } else {
        css = await fs__default.readFile(entry, "utf8");
        atImportOptions.root = path__default.dirname(entry);
        plugins.unshift(atImport__default(atImportOptions));
      }
      const res = await postcss__default(plugins).process(css, {
        from: void 0
      });
      return res;
    },
    processSync(entry) {
      const { sassOptions } = options;
      const plugins = this.getPluginsSync();
      let css;
      const extname = path__default.extname(entry);
      if (isExtSassFile(extname)) {
        try {
          css = sassCompileSync(entry, sassOptions);
        } catch (error) {
          throw error;
        }
      } else {
        css = fs__default$1.readFileSync(entry, "utf8");
      }
      const res = postcss__default(plugins).process(css).sync();
      return res;
    },
    generate: function() {
      const { generatorOptions } = options;
      return generator.generate(this, generatorOptions);
    }
  };
}

exports.createContext = createContext;
exports.getOptions = getOptions;
